package de.tu_darmstadt.tk.SmartHomeNetworkSim.core.protocols;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Packet;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Port;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Protocol;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.protocols.packets.MQTT_packet;

/**
 * Implementation of the MQTT Protocol to generate packets for the simulation
 * 
 *
 * @author Andreas T. Meyer-Berg
 */
public class MQTT_protocol implements Protocol {

	/**
	 * Broker which collects and distributes messages
	 */
	private Port broker;

	/**
	 * Publishers like sensors, which publish data
	 */
	private LinkedList<Port> pubs;

	/**
	 * Subscriber which subscribe to different Topics
	 */
	private LinkedList<Port> subs;
	
	/**
	 * Topics subscribed by each Port
	 */
	private HashMap<Port,LinkedList<String>> subbedTopics = new HashMap<Port,LinkedList<String>>();

	/**
	 * Devices that are Publisher and Subscriber and therefore send and receive
	 * messages
	 */
	private LinkedList<Port> pubSubs;

	/**
	 * Packets which are currently being generated or were generated after the
	 * last generatePackets Call
	 */
	private LinkedList<Packet> currentPackets = new LinkedList<Packet>();
	
	/**
	 * Topics of the MQTT Broker
	 */
	private LinkedList<String> topics = new LinkedList<String>();

	/**
	 * Creates a new MQTT Protocol
	 * 
	 * @param broker broker of the protocol
	 */
	public MQTT_protocol(Port broker) {
		this.broker = broker;
	    topics.add("/home/temperatureHot");
	    topics.add("/home/doorOpen");
	    topics.add("/home/lightOn");
	    subs = new LinkedList<Port>();
	    pubs = new LinkedList<Port>();
	    pubSubs = new LinkedList<Port>();
	    
	}
	
	@Override
	public Collection<Packet> generateNextPakets(Port port, long timestep, boolean packetLost) {
		/**
		 * Packets which will be generated
		 */
		LinkedList<Packet> returnPackets = new LinkedList<Packet>();
		//Update all Timestamps of previous created Packets
		for(Packet p: currentPackets){
			p.setTimestamp(p.getTimestamp()+timestep);
		}
		//Add these packets to the return
		returnPackets.addAll(currentPackets);
		//remove packets from the old list
		currentPackets.clear();
		
		//Return termination packets
		if(port==null)return returnPackets;
		
		/**
		 * Update the lastTime the port was triggered
		 */
		port.setLastTrigger(timestep);
		
		/**
		 * Generate new Packets regarding to their class
		 */
		if(port==broker){
			//Broker should not send new packages, without new messages
		}else if(subs.contains(port)||pubSubs.contains(port)){
			
			//Subs can either subscribe to topics or unsubscribe
			/**
			 * Topics, the SmartDevice is subscribed to
			 */
			LinkedList<String> tops= subbedTopics.get(port);
			
			/**
			 * Topic which should be subscribed or unsubscribed
			 */
			String newTopic;
			//Subsribe if no subscriptions so far or not subscribed to all, with a probability of70%
			if(tops.size()==0||(tops.size()!=subbedTopics.size()&&Math.random()<0.7)){
				//Forced Subscribe
				newTopic = topics.get((int) Math.floor(Math.random()*topics.size()));
				returnPackets.add(new MQTT_packet(MQTT_packet.SUBSCRIBE, timestep, port, broker, "topic:"+newTopic));
				timestep+=broker.getResponseTime();
				returnPackets.add(new MQTT_packet(MQTT_packet.SUBACK, timestep, broker, port));
				tops.add(newTopic);
			}
			else if(tops.size()==subbedTopics.size()){
				//Forced Unsubscribe
				newTopic = tops.get((int) Math.floor(Math.random()*tops.size()));
				returnPackets.add(new MQTT_packet(MQTT_packet.UNSUBSCRIBE, timestep, port, broker, "topic:"+newTopic));
				timestep+=broker.getResponseTime();
				returnPackets.add(new MQTT_packet(MQTT_packet.UNSUBACK, timestep, broker, port));
				tops.remove(newTopic);
			}
			
			if(pubSubs.contains(port)&&Math.random()<0.3){
				//When also Pub, publish sometimes 
				newTopic = topics.get((int) Math.floor(Math.random()*topics.size()));
				returnPackets.add(new MQTT_packet(MQTT_packet.SUBSCRIBE, timestep, port, broker, "Topic:"+newTopic));
				timestep+=broker.getResponseTime();
				returnPackets.add(new MQTT_packet(MQTT_packet.SUBACK, timestep, broker, port));
			}
			
		}
		if(pubs.contains(port)||pubSubs.contains(port)&&Math.random()<0.3){
			String newTopic = topics.get((int) Math.floor(Math.random()*topics.size()));
			/**
			 * Message to be published (Should be further specialized by a Sensor Device) 
			 */
			String msg = "Topic:"+newTopic+":"+(Math.random()<0.5?"true":"false");
			returnPackets.add(new MQTT_packet(MQTT_packet.PUBLISH, timestep, port, broker, msg));
			//Response
			timestep+=broker.getResponseTime();
			returnPackets.add(new MQTT_packet(MQTT_packet.PUBACK, timestep, broker, port));
			
			//Publish to Subscribers
			//Should be be improved to just notify Subs that are subscribed to the topic
			for(Port p:subs){
				if(!subbedTopics.get(p).contains(newTopic))continue;//Skip unsubbed ports
				timestep+=broker.getResponseTime();
				returnPackets.add(new MQTT_packet(MQTT_packet.PUBLISH, timestep, broker, p, msg));
				returnPackets.add(new MQTT_packet(MQTT_packet.PUBACK, timestep+p.getResponseTime(), p, broker));
			}
			for(Port p:pubSubs){
				if(!subbedTopics.get(p).contains(newTopic))continue;//skip unsubbed ports
				timestep+=broker.getResponseTime();
				returnPackets.add(new MQTT_packet(MQTT_packet.PUBLISH, timestep, broker, p, msg));
				returnPackets.add(new MQTT_packet(MQTT_packet.PUBACK, timestep+p.getResponseTime(), p, broker));
			}
		}
		return returnPackets;
	}

	@Override
	public int getNumberOfRoles() {
		return 4;
	}

	@Override
	public String[] getRoles() {
		// PublisherSubscriber is Publisher as well as Subscriber
		return new String[] { "Broker", "PublisherSubsriber", "Publisher", "Subscriber" };
	}

	@Override
	public Collection<Port> getDevicesWithRole(int role) {
		switch (role) {
		case 0:
			return new LinkedList<Port>(Arrays.asList(broker));
		case 1:
			return pubSubs;
		case 2:
			return pubs;
		case 3:
			return subs;

		default:
			return null;
		}
	}

	@Override
	public boolean addDeviceOfRole(Port device, int role) {
		/*
		 * First device has to be the Broker 
		 */
		if (broker==null) {
			if (role == 0) {
				broker = device;
				return true;
			} else {
				return false;
			}
		}
		switch (role) {
		case 0:
			//Just one broker allowed.
			return false;
		case 1:
			pubSubs.add(device);

			subbedTopics.putIfAbsent(device, new LinkedList<String>());
			break;
		case 2:
			pubs.add(device);
			break;
		case 3:
			subs.add(device);
			subbedTopics.putIfAbsent(device, new LinkedList<String>());
			break;
		default:
			// invalid role
			return false;
		}
		// Create packets for the connecting Client
		currentPackets.add(new MQTT_packet(MQTT_packet.CONNECT, currentPackets.size()/2, device, broker));
		currentPackets.add(new MQTT_packet(MQTT_packet.CONNACK, currentPackets.size()/2+1, broker, device));
		return true;
	}

	@Override
	public void removeDevice(Port device) {
		/**
		 * true if the device was removed
		 */
		boolean removedDevice=false;
		if (broker == device){
			broker = null;
		}
		removedDevice|=pubSubs.remove(device);
		removedDevice|=subs.remove(device);
		removedDevice|=pubs.remove(device);
		//Remove Port from topics and clear its list
		LinkedList<String> oldTopics = subbedTopics.remove(device);
		if(oldTopics!=null)oldTopics.clear();
		if(removedDevice && broker!=null){
			//If not the broker and device was removed -> disconnect
			currentPackets.add(new MQTT_packet(MQTT_packet.DISCONNECT, currentPackets.size()/2, device, broker));
		}
	}
	
	public void addTopic(String topic){
		topics.add(topic);
	}

}
