package de.tu_darmstadt.tk.SmartHomeNetworkSim.view;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.WindowEvent;
import java.awt.event.WindowStateListener;
import java.awt.image.BufferedImage;

import javax.imageio.ImageIO;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;

import java.awt.BorderLayout;
import java.io.File;
import java.io.InputStream;

import javax.swing.SwingConstants;

import de.tu_darmstadt.tk.SmartHomeNetworkSim.control.Controller;
import de.tu_darmstadt.tk.SmartHomeNetworkSim.core.Model;

/**
 * Main class of the GUI, which combines the UserInterface and visualisation of the simulation.
 *
 * @author Andreas T. Meyer-Berg
 */
@SuppressWarnings("serial")
public class MainFrame extends JFrame {
	
	public final VisualisationPanel panel; 
	
	/**
	 * Model of the smart home
	 */
	private Model model;
	private Controller control;
	
	/**
	 * Creates a new Frame for the program, which is the most outer frame of the application
	 * @param m Model which is represented
	 * @param c Controller which handles the user interaction
	 */
	public MainFrame(Model m, Controller c) {
		//Set model and control
		this.model = m;
		this.control = c;
		
		setTitle("Smart Home Network Simulator");
		
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		getContentPane().setLayout(new BorderLayout(0, 0));
		
		//Set initial size of the frame
		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		int width = screenSize.width; 
		int height = screenSize.height;
		this.setBounds(width/8, height/8, 6*width/8, 6*height/8);
		this.setMinimumSize(new Dimension(640, 480));
		
		panel = new VisualisationPanel(model, control);
		getContentPane().add(panel, BorderLayout.CENTER);
		//Update model Dimensions to visualisation size
		
		try {
			/**
			 * Icon of the MainFrame UI
			 */	
			String path = "images/SmartHomeNetworkSim_icon.jpeg";
			//this.setIconImage(ImageIO.read(new File("src/main/resources/images/SmartHomeNetworkSim_icon.jpeg")));
			this.setIconImage(loadFile(path));
		} catch (Exception e) {
			System.err.println("WARNING: Failed to load Icon image in MainFrame.java");
			e.printStackTrace();
		}
		
		JMenuBar menuBar = new JMenuBar();
		getContentPane().add(menuBar, BorderLayout.NORTH);
		
		JMenuItem mntmOptions = new JMenuItem("Options");
		mntmOptions.setHorizontalAlignment(SwingConstants.LEFT);
		menuBar.add(mntmOptions);
		
		this.setVisible(true);
		//Set Dimension of the model to the JPanel size
		control.setDimension(panel.getWidth(),panel.getHeight(),1000);	
		panel.delayedInit();
		
		//Update components if window was maximized
		this.addWindowStateListener(new WindowStateListener() {
			@Override
			public void windowStateChanged(WindowEvent e) {
				if ((e.getNewState() & MainFrame.MAXIMIZED_BOTH) == MainFrame.MAXIMIZED_BOTH){
					revalidate();
					panel.revalidate();
					panel.repaint();
				}
			}
		});
	}
	
	/**
	 * Loads resource image from the given path (resources/path).
	 * Return default image, if loading failed
	 * @param path pathe where the image is located
	 * @return Image from path
	 */
	public BufferedImage loadFile(String path){
		try {
			//Try loading Image from jar/resources
			InputStream i = this.getClass().getClassLoader().getResourceAsStream(path);
			return ImageIO.read(i);
		} catch (Exception e) {
			//Try loading from absolute path, if it failed (e.g. starting the program from eclipse or jvm) 
			//extend path with path to resources
			path = "src/main/resources/"+ path;
			try {
				return ImageIO.read(new File(path));
			} catch (Exception e2) {
				/**
				 * return default Image (red cross on white background)
				 */
				BufferedImage def = new BufferedImage(64, 64, BufferedImage.TYPE_INT_RGB);
				for(int i = 0; i<64; i++)
					for(int j = 0; j<64; j++){
						if(i*(64-i)-j*(64-j)==0)
							def.setRGB(i, j, Color.RED.getRGB());
						else
							def.setRGB(i, j, Color.white.getRGB());
					}
				return def;
			}
		}
	}
}
